import html
import json
import os
import random
import re
import time
from typing import Dict, List, Union, cast
from urllib.parse import urljoin

import requests
from authlib.integrations.requests_client import OAuth2Auth
from locust import HttpUser
from locust.clients import ResponseContextManager as Response
from pydantic import BaseModel

from performance_tests.common import TestData

# disable deprecation warning
# Certificate for login.kc1.broker.intranet has no `subjectAltName`,
# falling back to check for a `commonName` for now.
requests.packages.urllib3.disable_warnings()

SD_METADATA = "/ucsschool/apis/self_disclosure/v1/users/{id}/metadata"
SD_GROUPS = "/ucsschool/apis/self_disclosure/v1/users/{id}/groups"
SD_USERS = "/ucsschool/apis/self_disclosure/v1/groups/{id}/users"

UCS_ENV_TEST_APP_FQDN = os.environ.get("UCS_ENV_TEST_APP_FQDN")
LOCUST_USERS = os.environ.get("LOCUST_USERS", "2")

if not UCS_ENV_TEST_APP_FQDN:
    raise ValueError("ENV not set, check: UCS_ENV_TEST_APP_FQDN (e.g. kc1.broker.intranet)")

TESTAPP_URL = f"https://{UCS_ENV_TEST_APP_FQDN}/univention-test-app/"
TESTAPP_PARAM = "login?kc_idp_hint={idp_hint}&pkce=y"


class LoginError(Exception):
    """Custom error for everything Login related"""


class TestAppClient:
    @staticmethod
    def _extract_relay_state(page):
        relay_state = re.search('name="RelayState" value="([^"]+)"', page.content.decode("utf-8"))
        if relay_state is None:
            raise LoginError("could not extract relay state")
        return relay_state.group(1)

    @staticmethod
    def _extract_saml_msg(page):
        saml_message = re.search('name="SAMLRequest" value="([^"]+)"', page.content.decode("utf-8"))
        if saml_message is None:
            raise LoginError("could not extract saml message\n%s" % page.text)
        return saml_message.group(1)

    @staticmethod
    def _extract_saml_response(page):
        saml_response = re.search('name="SAMLResponse" value="([^"]+)"', page.content.decode("utf-8"))
        if saml_response is None:
            raise LoginError("could not extract saml response")
        return saml_response.group(1)

    @staticmethod
    def _extract_kc_url(page):
        url = re.search('form method="post"[ \r\n]* action="([^"]+)"', page.content.decode("utf-8"))
        if url is None:
            error = re.search("Nutzername oder Passwort falsch", page.content.decode("utf-8"))
            if error is None:
                raise LoginError("could not extract keycloak url")
            else:
                raise LoginError("invalid user")
        return url.group(1)

    @staticmethod
    def _extract_redirect_url(page):
        url = re.search('method="post" action="([^"]+)"', page.content.decode("utf-8"))
        if url is None:
            raise LoginError("could not extract redirect url")
        return url.group(1)

    @staticmethod
    def _extract_auth_state(page):
        auth_state = re.search('name="AuthState" value="([^"]+)"', page.content.decode("utf-8"))
        if not auth_state:
            raise LoginError("could not extract auth state")
        return auth_state.group(1)

    @staticmethod
    def _extract_access_token(page):
        token = re.search(
            '(<h2>Token endpoint response:</h2>\n<pre style="white-space: pre-wrap; word-wrap: break-word;">)'
            + "(.*)(</pre><br>\n<h2>Encoded access Token:</h2>)",
            page.content.decode("utf-8"),
            flags=re.DOTALL,
        )

        if token is None:
            raise LoginError("could not extract token")

        token = token.group(2)
        token = html.unescape(token)
        token = token.replace("'", '"')
        token = json.loads(token)

        return token

    @staticmethod
    def _extract_pseudonym(page):
        pseudonym = re.search("(<p id=pseudonym>)(.*)(</p>)", page.content.decode("utf-8"))
        if not pseudonym:
            raise LoginError("could not extract auth state")
        return pseudonym.group(2)

    @classmethod
    def get_access_token_and_pseudonym(cls, username, password, idp_hint):
        session = requests.Session()
        param = TESTAPP_PARAM.format(idp_hint=idp_hint)
        url = urljoin(TESTAPP_URL, param)
        page = session.get(url, verify=False)

        saml_msg = cls._extract_saml_msg(page)
        relay_state = cls._extract_relay_state(page)
        redirect_url = cls._extract_redirect_url(page)
        page = session.post(
            redirect_url, data={"SAMLResponse": saml_msg, "RelayState": relay_state}, verify=False
        )

        auth_state = cls._extract_auth_state(page)
        data = {"username": username, "password": password, "AuthState": auth_state}
        page = session.post(page.url, data=data, verify=False)

        kc_url = cls._extract_kc_url(page)
        saml_msg = cls._extract_saml_response(page)
        relay_state = cls._extract_relay_state(page)
        page = session.post(kc_url, data={"SAMLResponse": saml_msg, "RelayState": relay_state})

        token = cls._extract_access_token(page)
        pseudonym = cls._extract_pseudonym(page)

        return token, pseudonym

    @classmethod
    def refresh_token(cls, token):
        url = (
            f"https://login.{UCS_ENV_TEST_APP_FQDN}/auth/realms/ID-Broker/protocol/openid-connect/token"
        )
        data = {
            "client_id": "univention-test-app",
            "client_secret": "3oGGuLhyxiYVW3K6R7qm",
            "grant_type": "refresh_token",
            "refresh_token": token["refresh_token"],
        }
        page = requests.post(url, data=data, verify=False)
        try:
            token = json.loads(page.text)
        except json.JSONDecodeError as exc:
            raise LoginError("could not decode token") from exc
        return token


class ResponseError(Exception):
    ...


class SDMetadata(BaseModel):
    user_id: str
    username: str
    firstname: str
    lastname: str
    type: str

    school_authority: str
    school_id: str

    def __str__(self):
        data = self.__dict__
        del data["firstname"]
        del data["lastname"]
        return f"{self.__class__.__name__}{data}"


class SDUserGroups(BaseModel):
    groups: List[Dict[str, Union[str, int]]]

    def __str__(self):
        return f"{self.__class__.__name__}{self.groups}"


class SDGroupUsers(BaseModel):
    students: List[Dict[str, str]]
    teachers: List[Dict[str, str]]

    def __str__(self):
        str_students = self.students.copy()
        for student in str_students:
            del student["firstname"]
            del student["lastname"]
        return f"{self.__class__.__name__}(students: {str_students}, teachers: {self.teachers})"


class SelfDisclosureClient(HttpUser):
    abstract = True

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.username = None
        self.password = None
        self.idp_hint = None
        self._token = None
        self.pseudonym = None
        self._auth: OAuth2Auth = None

    @property
    def token(self):
        return self._token

    @token.setter
    def token(self, token):
        self._token = token
        self._auth = OAuth2Auth(token, token_placement="header")

    @property
    def auth(self):
        return self._auth

    def _check_get_data(self, response: Response):
        msg: str = f"status code: {response.status_code} Response text: {response.text}"
        if response.status_code == 401:
            response.failure(msg)
            token = TestAppClient.refresh_token(self.token)
            self.token = token
            raise ResponseError(msg, " -> Token Refresh")
        elif response.status_code != 200:
            response.failure(msg)
            raise ResponseError(msg)
        try:
            data = json.loads(response.text)
        except json.JSONDecodeError as exc:
            raise ResponseError(msg) from exc
        return data

    def get_user_metadata(self):
        url: str = f"https://{self.host}{SD_METADATA.format(id=self.pseudonym)}"
        with self.client.get(
            url, name=SD_METADATA, auth=self.auth, verify=False, catch_response=True
        ) as response:
            response = cast(Response, response)
            metadata = self._check_get_data(response)
        return SDMetadata(school_authority=self.idp_hint, **metadata)

    def get_groups_of_user(self):
        url: str = f"https://{self.host}{SD_GROUPS.format(id=self.pseudonym)}"
        with self.client.get(
            url, name=SD_GROUPS, auth=self.auth, verify=False, catch_response=True
        ) as response:
            response = cast(Response, response)
            groups = self._check_get_data(response)
        return SDUserGroups(groups=groups["groups"])

    def get_users_of_group(self, id):
        url: str = f"https://{self.host}{SD_USERS.format(id=id)}"
        with self.client.get(
            url, name=SD_USERS, auth=self.auth, verify=False, catch_response=True
        ) as response:
            response = cast(Response, response)
            users = self._check_get_data(response)
        return SDGroupUsers(students=users["students"], teachers=users["teachers"])


def get_student_data() -> List[Dict[str, str]]:
    return _get_user_data("student")


def get_teacher_data(get_groups=False) -> List[Dict[str, str]]:
    return _get_user_data("teacher", get_groups)


def _get_user_data(get_role: str, get_groups=False) -> List[Dict[str, str]]:
    locust_users = int(LOCUST_USERS)
    numbers = {
        5: (2, 3, 1),
        10: (2, 3, 2),
        15: (2, 4, 2),
        20: (2, 5, 2),
    }

    sa_no: int  # number of school authorities to load
    sch_no: int  # number of schools to load per school authority
    user_no: int  # number of users to load per school

    if locust_users <= 20:
        for i in range(5, 21, 5):
            if locust_users <= i:
                sa_no, sch_no, user_no = numbers[i]
                break
    else:
        k = ((locust_users // 10 * 10) + ((locust_users % 10 > 0) * 10)) // 10
        sa_no, sch_no, user_no = (2, 5, k)

    td = TestData()
    data: List[Dict[str, str]] = list()
    users = list()
    for sa in random.choices(td.school_authorities, k=sa_no):
        for school in random.choices(list(td.school_authority(sa).schools.keys()), k=sch_no):
            school_data = td.school(sa, school)
            students = list()
            teachers = list()
            for user in school_data.users:
                if "student" in school_data.users[user].ucsschoolRole:
                    students.append((school_data.users[user], school, sa, school_data.groups))
                elif "teacher" in school_data.users[user].ucsschoolRole:
                    teachers.append((school_data.users[user], school, sa, school_data.groups))

            if get_role == "student":
                users += random.choices(students, k=user_no)
            elif get_role == "teacher":
                users += random.choices(teachers, k=user_no)

    print(f"\nRetrieve tokens for {len(users)} users:")

    t_start = time.time()

    for user_data in users:
        user = user_data[0]
        school = user_data[1]
        sa = user_data[2]
        groups = user_data[3]
        user_groups = list()
        if get_groups:
            for group in groups:
                if user.name in groups[group].memberUid:
                    user_groups.append(groups[group].pseudonyms["idBrokerPseudonym0001"])
        username = f"{school}-{user.name}"
        password = user.password
        role = user.ucsschoolRole
        idp_hint = sa.lower()
        print(f"  user: {user.name}, sa: {sa}, school: {school}, role: {role}")
        token, pseudonym = TestAppClient.get_access_token_and_pseudonym(username, password, idp_hint)
        groups = None
        entry = {
            "username": username,
            "password": password,
            "role": role,
            "idp_hint": idp_hint,
            "token": token,
            "pseudonym": pseudonym,
            "groups": user_groups,
        }
        data.append(entry)
    print(f"time elapsed: {round((time.time() - t_start), 2)} seconds")
    return data
